#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from turberfield.catchphrase.presenter import Presenter
from turberfield.catchphrase.render import Action
from turberfield.catchphrase.render import Parameter
from turberfield.catchphrase.render import Renderer
from turberfield.catchphrase.render import Settings
from turberfield.dialogue.types import DataObject


class Story(Renderer, DataObject):

    def __init__(self, cfg=None, context=None, **kwargs):
        super().__init__(**kwargs)

        self.definitions = {
            "catchphrase-colour-washout": "hsl(50, 0%, 100%, 1.0)",
            "catchphrase-colour-shadows": "hsl(202.86, 100%, 4.12%)",
            "catchphrase-colour-midtone": "hsl(203.39, 96.72%, 11.96%)",
            "catchphrase-colour-hilight": "hsl(203.06, 97.3%, 56.47%)",
            "catchphrase-colour-glamour": "hsl(353.33, 96.92%, 12.75%)",
            "catchphrase-colour-gravity": "hsl(293.33, 96.92%, 12.75%)",
            "catchphrase-reveal-extends": "both",
        }
        self.settings = Settings(**self.definitions)
        self.drama = {} if not context else [context]

    @property
    def actions(self):
        yield Action(
            "cmd", None, "/{0.id.hex}/cmd/", [self], "post",
            [Parameter("cmd", False, self.context.validator, [self.context.prompt], ">")],
            "Enter"
        )

    @property
    def context(self):
        return next(iter(self.drama), None)

    def refresh_target(self, url):
        refresh_state = getattr(self.settings, "catchphrase-states-refresh", "inherit").lower()
        if refresh_state == "none":
            return None
        elif refresh_state == "inherit":
            return url
        else:
            return refresh_state

    def represent(self, *args, facts=None, previous=None, factory=Presenter, **kwargs):
        rv = self.context.interlude(
            self.context.folder,
            previous and previous.index,
            previous and previous.ensemble
        )
        presenter = factory.build_presenter(
            self.context.folder, *args, facts=facts or rv,
            ensemble=self.context.ensemble + [self.context, self.settings],
            **kwargs
        )
        if presenter and not(presenter.dwell or presenter.pause):
            setattr(self.settings, "catchphrase-reveal-extends", "none")
            setattr(self.settings, "catchphrase-states-scrolls", "scroll")
        else:
            setattr(self.settings, "catchphrase-reveal-extends", "both")
            setattr(self.settings, "catchphrase-states-scrolls", "visible")

        return presenter



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random
import unittest

from balladeer import Fruition
from balladeer import Gesture
from balladeer import Hand
from balladeer import Head
from balladeer import Phrase
from balladeer import Name
from balladeer import Verb


class Brew(Gesture):

    @staticmethod
    def create_mugs(a, b):
        return Gesture(
            "mugs",
            a=a, b=b,
            head=Head(
                propose=["Can you get the mugs for me?"],
                confirm=["OK, fine."],
                counter=["Don't worry, I'll do it."],
                abandon=["Oh, never mind."],
                condemn=["There's a crack in that one.", "That needs a wash."],
                declare=["Right then."],
            ),
            # TODO: Add abandon
            hand=Hand(
                decline=["I can't right now."],
                suggest=["I will in a minute."],
                promise=["Sure."],
                disavow=["I though you were doing it."],
                deliver=["There they are."],
            ),
        ).set_state(Fruition.inception)

    @staticmethod
    def strategy(options):
        return random.choice(
            [(e, s) for e, s in options
            if s not in (Fruition.cancelled, Fruition.defaulted, Fruition.withdrawn)]
        )

    def __init__(self, *args, mugs=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.mugs = mugs or self.create_mugs(a=getattr(self, "b", None), b=getattr(self, "a", None))

    def __call__(self, strategy=None, **kwargs):
        if self.mugs.failed:
            self.mugs = self.create_mugs(a=getattr(self, "b", None), b=getattr(self, "a", None))

        if self.get_state(Fruition) == Fruition.construction and not self.mugs.passed:
            g, e, s = self.mugs(strategy=self.strategy)
            if e == self.mugs.hand.suggest:
                self.mugs.b = self.mugs.a
            return g, e, s

        return super().__call__(strategy, **kwargs)


class GestureTests(unittest.TestCase):

    @staticmethod
    def create_brew():
        return Brew(
            "brew",
            a="Louise", b="Sophie",
            head=Head(
                propose=["Stick the kettle on, would you?"],
                confirm=["Whatever, fine."],
                counter=["I'll have tea, please."],
                abandon=["Actually, forget it; I've got to go."],
                condemn=["You've left the bag in."],
                declare=["Thanks."],
            ),
            hand=Hand(
                decline=["No time now, sorry."],
                suggest=["Coffee?"],
                promise=["OK then."],
                disavow=["I'll be back in a minute. Carry on without me."],
                deliver=["There you go."],
            ),
        ).set_state(Fruition.inception)

    def test_head(self):
        a = Head()
        b = Head()
        self.assertFalse(a.propose)
        self.assertFalse(b.propose)
        self.assertIs(a.propose, b.propose)  # Lest we forget

    def test_simple(self):
        g = Gesture("simple", head=Head([
            Phrase(Verb("make"), Name("tea")),
            Phrase(Verb("make"), Name("brew")),
        ]))
        self.assertEqual("simple", str(g))
        self.assertEqual("make", g.head.propose[0].verb.imperative)
        self.assertEqual("tea", g.head.propose[0].name.noun)

    def test_simple(self):
        g = Gesture("simple", head=Head([
            Phrase(Verb("make"), Name("tea")),
            Phrase(Verb("make"), Name("brew")),
        ]))
        g.propose = "Fancy a brew?"
        self.assertEqual("Fancy a brew?", g.propose)
        self.assertEqual("make", g.head.propose[0].verb.imperative)
        self.assertEqual("tea", g.head.propose[0].name.noun)
        g.decline = "I've just had one, thanks"
        self.assertEqual("I've just had one, thanks", g.decline)
        self.assertFalse(g.hand.decline)

    def test_attribute_access(self):
        g = Gesture("blank")
        self.assertRaises(AttributeError, getattr, g, "new_attribute")

        self.assertFalse(g.propose)
        self.assertFalse(g.decline)
        g.new_attribute = True
        self.assertTrue(g.new_attribute)

    def test_brew(self):
        brew = self.create_brew()
        while not (brew.passed or brew.failed):
            gesture, event, state = brew()
            gesture.state = state


if __name__ == "__main__":
    """
    Louise : Stick the kettle on, would you?
    Sophie : Coffee?
    Sophie : OK then.
    Sophie : Can you get the mugs for me?
    Louise : Sure.
    Louise : There they are.
    Sophie : There's a crack in that one.
    Louise : There they are.
    Sophie : Right then.
    Sophie : There you go.
    Louise : You've left the bag in.
    Sophie : I'll be back in a minute. Carry on without me.
    """
    brew = GestureTests.create_brew()

    while not (brew.passed or brew.failed):
        gesture, event, state = brew()
        gesture.state = state

        if event in brew.head:
            actor = brew.a
        elif event in brew.hand:
            actor = brew.b
        elif event in brew.mugs.head:
            actor = brew.mugs.a
        elif event in brew.mugs.hand:
            actor = brew.mugs.b

        print(actor, ":", random.choice(event))



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import textwrap
import unittest

from balladeer import Drama
from balladeer import Presenter


class DramaTests(unittest.TestCase):

    def test_next_states_single(self):
        text = textwrap.dedent("""
        .. entity:: DRAMA
           :types:  balladeer.Drama

        Scene
        =====

        Shot
        ----

        .. condition:: DRAMA.state 0

        Text.
        """)
        drama = Drama()
        presenter = Presenter.build_from_text(text, ensemble=[drama])
        self.assertTrue(presenter.frames, presenter.text)

        drama.valid_states = drama.find_valid_states(presenter)
        self.assertEqual([0], drama.valid_states)

        rv = drama.next_states()
        self.assertEqual((0, 0), rv)

    def test_next_states_gaps(self):
        text = textwrap.dedent("""
        .. entity:: DRAMA
           :types:  balladeer.Drama

        Scene
        =====

        Zero
        ----

        .. condition:: DRAMA.state 0

        Text.

        One
        ---

        .. condition:: DRAMA.state 1

        Text.

        Four
        ----

        .. condition:: DRAMA.state 4

        Text.

        Five
        ----

        .. condition:: DRAMA.state 5

        Text.
        """)
        drama = Drama()
        presenter = Presenter.build_from_text(text, ensemble=[drama])
        self.assertTrue(presenter.frames, presenter.text)

        drama.valid_states = drama.find_valid_states(presenter)
        self.assertEqual([0, 1, 4, 5], drama.valid_states)

        drama.state = 0
        rv = drama.next_states()
        self.assertEqual((0, 1), rv)

        drama.state = 1
        rv = drama.next_states()
        self.assertEqual((0, 4), rv)

        drama.state = 4
        rv = drama.next_states()
        self.assertEqual((1, 5), rv)

        drama.state = 5
        rv = drama.next_states()
        self.assertEqual((4, 5), rv)



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
import enum
import random

from turberfield.dialogue.types import DataObject
from turberfield.dialogue.types import EnumFactory
from turberfield.utils.misc import group_by_type

from balladeer.speech import Name


class Fruition(EnumFactory, enum.Enum):
    inception = 1
    elaboration = 2
    construction = 3
    transition = 4
    completion = 5
    discussion = 6
    defaulted = 7
    withdrawn = 8
    cancelled = 9


class Grouping(defaultdict):

    @property
    def all(self):
        return [i for s in self.values() for i in s]

    @property
    def each(self):
        return list(set(self.all))


class Named(DataObject):

    @property
    def name(self):
        name = random.choice(getattr(self, "names", [Name()]))
        article = name.article.definite and f"{name.article.definite} "
        return f"{article}{name.noun}"

    def __str__(self):
        return "\n".join(i.noun for i in self.names)


class Operation(enum.Enum):

    begins = enum.auto()
    frames = enum.auto()
    paused = enum.auto()
    prompt = enum.auto()
    ending = enum.auto()
    finish = enum.auto()


class World:

    def __init__(self, *args, **kwargs):
        self.lookup = Grouping(list)
        for item in self.build():
            self.add(item)

    @staticmethod
    def arrange(items):
        return group_by_type(items)

    def add(self, item):
        for name in str(item).splitlines():
            self.lookup[name.strip().lower()].append(item)

    def remove(self, item):
        for name in str(item).splitlines():
            try:
                self.lookup[name.strip().lower()].remove(item)
            except ValueError:
                pass



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cmath
from collections import defaultdict
import enum

from turberfield.dialogue.types import DataObject
from turberfield.dialogue.types import EnumFactory
from turberfield.dialogue.types import Stateful
from turberfield.utils.homogeneous import vector


class Compass(EnumFactory, enum.Enum):

    N   = vector(+0, +1)
    NE  = vector(+1, +1)
    E   = vector(+1, +0)
    SE  = vector(+1, -1)
    S   = vector(+0, -1)
    SW  = vector(-1, -1)
    W   = vector(-1, +0)
    NW  = vector(-1, -1)

    @property
    def back(self):
        return {
            "N":  Compass.S,
            "NE": Compass.SW,
            "E":  Compass.W,
            "SE": Compass.NW,
            "S":  Compass.N,
            "SW": Compass.NE,
            "W":  Compass.E,
            "NW": Compass.SE
        }.get(self.name)

    @property
    def bearing(self):
        phase = 180 * cmath.phase(complex(*self.value[:2])) / cmath.pi
        if phase <= 90:
            rv = 90 - phase
        elif phase <= 180:
            rv = 270 + (180 - phase)
        elif phase <= 270:
            rv = 180 + (270 - phase)
        else:
            rv = 90 + 360 - phase
        return rv % 360


class Waypoint:

    @property
    def label(self):
        return self.value[0]

    @property
    def title(self):
        return self.label.title()


class Via(EnumFactory, enum.Enum):
    block = 0
    forwd = 1
    bckwd = 2
    bidir = 3


class Transit(DataObject, Stateful): pass


class Map:

    def __init__(self, exit:type, into:type, **kwargs):
        self.exit = exit or getattr(self, "Exit")
        self.into = into or getattr(self, "Into")
        self.transits = []
        self.routes = {}

    @property
    def topology(self):
        for t in self.transits:
            d = t.get_state(self.exit)
            a = t.get_state(self.into)
            v = t.get_state(Via)
            c = t.get_state(Compass)
            b = c and c.back
            if v in (Via.bidir, Via.forwd):
                yield d, c, t, a
            if v in (Via.bidir, Via.bckwd):
                yield a, b, t, d

    def options(self, waypoint):
        typ = type(waypoint)
        return {
            (c or n, typ[a.name], t)
            for n, (d, c, t, a) in enumerate(self.topology)
            if d.name == waypoint.name
        }

    def route(self, locn, dest):
        if (locn.name, dest.name) in self.routes:
            return self.routes[(locn.name, dest.name)]

        rvs = set()
        paths = [[locn.name]]

        graph = defaultdict(set)
        for d, _, t, a in self.topology:
            graph[d.name].add(a.name)

        n = len(graph)
        d = 1
        while n >= 0 or not rvs:
            nxt = []
            for p in paths:
                if p[-1] == dest.name:
                    rvs.add(tuple(p))
                else:
                    nodes = graph[p[-1]]
                    d = len(nodes)
                    for i in nodes:
                        if i not in p:
                            nxt.append(p.copy())
                            nxt[-1].append(i)
            paths = nxt
            n = n - d

        rv = [type(locn)[i] for i in sorted(rvs, key=len)[0]] if rvs else []
        self.routes[(locn.name, dest.name)] = rv
        return rv



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple

from turberfield.dialogue.types import DataObject
from turberfield.dialogue.types import Stateful


Article = namedtuple(
    "Article",
    ("definite", "indefinite"),
    defaults=("the", "a")
)


Pronoun = namedtuple(
    "Pronoun",
    ("subject", "object", "reflexive", "genitive"),
    defaults=("it", "it", "itself", "its")
)


Name = namedtuple(
    "Name",
    ("noun", "article", "pronoun"),
    defaults=("", Article(), Pronoun())
)


Tensed = namedtuple("Tensed", ("simple", "progressive", "perfect", "imperative"))


class Verb(Tensed):

    def __new__(cls, root, simple="{0}s", progressive="is {0}ing", perfect="{0}ed", imperative="{0}"):
        l = locals()
        return super().__new__(cls, *(l[i].format(root) for i in super()._fields))


Phrase = namedtuple("Phrase", ("verb", "name"))



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import unittest

from balladeer.cartography import Compass
from balladeer.cartography import Map
from balladeer.cartography import Transit
from balladeer.cartography import Via
from balladeer.cartography import Waypoint


class CompassTests(unittest.TestCase):

    def test_nearing(self):
        self.assertEqual(0, Compass.N.bearing)
        self.assertEqual(180, Compass.S.bearing)

    def test_back(self):
        self.assertEqual(Compass.SW, Compass.NE.back)
        self.assertEqual(Compass.NE, Compass.SW.back)

    def test_vectors(self):
        self.assertEqual(Compass.NE.value, Compass.N.value + Compass.E.value)


class MapTests(unittest.TestCase):

    class SimpleMap(Map):

        spots = {
            "bedroom":  ["bedroom"],
            "hall":  ["hall", "hallway"],
            "kitchen":  ["kitchen"],
            "stairs":  ["stairs", "stairway", "up", "up stairs", "upstairs"],
            "inventory":  None
        }

        Into = enum.Enum("Into", spots, type=Waypoint)
        Exit = enum.Enum("Exit", spots, type=Waypoint)
        Spot = enum.Enum("Spot", spots, type=Waypoint)

        def __init__(self, exit=None, into=None, **kwargs):
            super().__init__(exit, into, **kwargs)
            exit = self.exit
            into = self.into
            self.transits = [
                Transit(label="bedroom door").set_state(exit.bedroom, into.hall, Via.bidir),
                Transit().set_state(exit.hall, Compass.N, into.stairs, Via.bidir),
                Transit(label="kitchen door").set_state(exit.kitchen, Compass.SW, into.hall, Via.bidir),
            ]

    def test_simple_options(self):
        m = MapTests.SimpleMap()
        self.assertEqual(3, len(m.options(m.Spot.hall)))

    def test_simple_route(self):
        m = MapTests.SimpleMap()
        r = m.route(m.Spot.kitchen, m.Spot.bedroom)
        self.assertEqual(3, len(r))
        self.assertEqual(3, len(set(r)))



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bisect
import re

from turberfield.catchphrase.mediator import Mediator
from turberfield.dialogue.model import Model
from turberfield.dialogue.types import Stateful


class Drama(Stateful, Mediator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validator = re.compile("[\\w ]+")
        self.prompt = "?"
        self.input_text = ""
        self.default_fn = None
        self.valid_states = [0]

    @property
    def ensemble(self):
        return []

    @property
    def turns(self):
        return len(self.history)

    def __call__(self, fn, *args, **kwargs):
        text, presenter, *_ = args
        # FIXME: Bring in Operation and test for paused.
        if presenter and (presenter.dwell or presenter.pause):
            self.valid_states = self.find_valid_states(presenter)

        return super().__call__(fn, *args, **kwargs)

    def build(self):
        return []

    def find_valid_states(self, presenter):
        return sorted(
            c.value for f in presenter.frames for c in f[Model.Condition]
            if c.object is self and c.format == "state" and isinstance(c.value, int)
        ) or [0]

    def next_states(self, n=1):
        fwd = min(bisect.bisect_right(self.valid_states, self.state) + n - 1, len(self.valid_states) - 1)
        bck = max(0, bisect.bisect_left(self.valid_states, self.state) - n)
        rv = (self.valid_states[bck], self.valid_states[fwd])
        return rv

    def interlude(self, folder, index, *args, **kwargs):
        return {}

    def deliver(self, cmd, presenter):
        self.input_text = cmd
        fn, args, kwargs = self.interpret(self.match(cmd, context=presenter, ensemble=self.ensemble))
        fn = fn or self.default_fn
        return fn and self(fn, *args, **kwargs)



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2022 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
TODO: Generate orders as per turberfield-dialogue:main.

Usage::

    python -m tools.folio blathnaid/tale/ | weasyprint - folio.pdf

"""

import argparse
import datetime
import mimetypes
import pathlib
import sys

from balladeer import Drama
from balladeer import Story
from turberfield.dialogue.adapters import ColourAdapter
from turberfield.dialogue.cli import add_common_options
from turberfield.dialogue.cli import add_performance_options
from turberfield.dialogue.main import HTMLHandler
from turberfield.dialogue.model import Model
from turberfield.utils.logger import LogAdapter
from turberfield.utils.logger import LogManager

import textwrap

# TODO
#  Format images
#  Prepend sections from .html files
#  Output is a directory, ?.html + images


class Folio(Story):

    static_style = textwrap.dedent("""
        @page {
            size: 5in 7.75in;
            counter-reset: footnote;
            @bottom-center {
                content: counter(page);
                width: 100%;
                vertical-align: top;
                border-top: 0.5pt solid;
                margin-top: 0.25in;
                margin-bottom: 0.6in;
                padding-top: 0.1in;
            }
            @footnote {
                footnote-display: inline;
                max-height: 2rem;
            }

        }

        @page:left{
            margin: 0.6in 0.75in 0.6in 0.5in;
            @top-left {
                content: var(--balladeer-metadata-project);
                font-size: 0.7rem;
                vertical-align: middle;
                width: 100%;
            }
        }

        @page:right{
            margin: 0.6in 0.5in 0.6in 0.75in;
            @top-right {
                content: var(--balladeer-metadata-version);
                font-size: 0.7rem;
                vertical-align: middle;
                width: 100%;
            }
        }

        @media print {
        .shot h2 {
        display: none;
        }

        .scene {
        page-break-before: right;
        }

        .shot {
        page-break-inside: avoid;
        }

        .shot:first-of-type p {
        text-indent: 0rem;
        }

        h1 {
        font-size: 200%;
        margin-bottom: 18%;
        margin-top: 28%;
        text-transform: capitalize;
        text-align: center;
        }

        blockquote header {
        display: none;
        }

        p {
        font-family: "Libre Baskerville", serif;
        font-size: small;
        margin-bottom: 1.3rem;
        text-align: justify;
        text-indent: 0.5rem;
        }

        .line:last-of-type blockquote::after {
        content: "-";
        display: block;
        margin-bottom: 1rem;
        margin-top: 1rem;
        text-align: center;
        }

        .spoken + .unspoken .line:first-of-type p {
        text-indent: 0rem;
        }

        blockquote {
        font-style: italic;
        }

        dl {
            display: none;
        }

        span.call{
        display: inline;
        }

        span.footnote {
        float: footnote;
        font-family: "Libre Baskerville", serif;
        font-size: 0.7rem;
        }

        ::footnote-call {
        display: inline;
        }

        ::footnote-marker {
        }
        }   /* End of print media */

        @media screen {
        dl {
        align-items: center;
        border-color: gray;
        border-top-style: dotted;
        border-width: thin;
        display: flex;
        flex-direction: row;
        flex-flow: wrap;
        flex-wrap: wrap;
        font-family: "DejaVu Sans", sans-serif;
        justify-content: space-around;
        letter-spacing: 0.125rem;
        padding-top: 1.6rem;
        margin-left: 1rem;
        }

        dt {
        font-size: 0.7rem;
        font-weight: lighter;
        margin-bottom: 0.4rem;
        margin-right: 0.3rem;
        }

        dt::after {
        content: ":";
        }

        dd {
        font-weight: lighter;
        margin-bottom: 0.9rem;
        margin-right: 1.0rem;
        }

        h1 {
        /* font-family: "Cinzel", "Libre Baskerville", serif; */
        font-size: 1.2rem;
        font-weight: bold;
        letter-spacing: 0.125rem;
        margin-top: 1rem;
        margin-bottom: 1rem;
        text-align: center;
        text-transform: uppercase;
        }

        h2 {
        border-color: gray;
        border-left-style: dotted;
        border-width: thin;
        font-family: "DejaVu Sans", sans-serif;
        font-size: 0.7rem;
        padding-left: 0.4rem;
        margin-bottom: 0.6rem;
        margin-left: 24rem;
        }

        section {
        margin-left: auto;
        margin-right: auto;
        margin-top: 6.0rem;
        width: 32rem;
        }

        .scene {
        margin-top: 1.8rem;
        }

        .shot {
        margin-bottom: 1.6rem;
        page-break-inside: avoid;
        }

        p {
        font-family: "IBM Plex Serif", "Junge", "Libre Baskerville", "FreeSerif", serif;
        line-height: 1.6rem;
        padding-bottom: 0.4rem;
        padding-top: 0.4rem;
        word-spacing: 0.3rem;
        }

        .spoken p {
        font-style: italic;
        }

        .spoken .line:first-of-type header {
        display: block;
        font-family: "DejaVu Sans", sans-serif;
        margin-bottom: 0.6rem;
        }

        blockquote header {
        display: none;
        }

        }   /* End of screen media */

        * {
        box-sizing: border-box;
        border: 0;
        font: inherit;
        font-size: 100%;
        line-height: 1.2em;
        list-style-type: none;
        margin: 0;
        outline: 0;
        padding: 0;
        text-decoration: none;
        vertical-align: baseline;
        }

        .catchphrase-banner {
        display: none;
        }

        em {
        font-style: italic;
        }

        strong {
        font-weight: bold;
        }

        .line:first-of-type blockquote header {
        text-transform: uppercase;
        }

    """)

    def __init__(self, dwell, pause, **kwargs):
        super().__init__(**kwargs)
        self.dwell = dwell
        self.pause = pause
        self.chapters = []
        self.metadata = {
            "print": datetime.date.today()
        }
        self.sections = []
        self.seconds = 0
        self.log = LogManager().get_logger("main").clone("folio")

    def animated_line_to_html(self, anim, **kwargs):
        name = getattr(anim.element.persona, "name", anim.element.persona)
        name = "{0.firstname} {0.surname}".format(name) if hasattr(name, "firstname") else name
        delay = self.seconds
        yield f'<div class="line" style="animation-delay: {delay:.2f}s; animation-duration: {anim.duration:.2f}s">'
        if name:
            yield "<blockquote>"
            yield f"<header>{name}</header>"
            yield f"{anim.element.html}".rstrip()
            yield "</blockquote>"
        else:
            yield f"{anim.element.html}".rstrip()
        yield "</div>"
        self.seconds += anim.duration

    def render_animated_frame_to_html(self, frame, controls=[], **kwargs):
        witness = next(i.element for v in frame.values() for i in v if hasattr(i, "element"))
        dialogue = "\n".join(i for l in frame[Model.Line] for i in self.animated_line_to_html(l, **kwargs))
        stills = "\n".join(self.animated_still_to_html(i, **kwargs) for i in frame[Model.Still])
        spoken = any(anim.element.persona for anim in frame[Model.Line])
        chapter = ([i.get("chapter", 0) for i in self.chapters] or [0])[-1]
        if not self.chapters or self.chapters[-1].get("scene") != frame["scene"]:
            if chapter:
                yield "</section>"
            metadata = {"path": witness.path}
            if frame["scene"]:
                metadata.update({"chapter": chapter + 1, "scene": frame["scene"]})
            self.chapters.append(metadata)
            yield from self.render_metadata(**metadata)

            yield '<section class="scene">'

            if frame["scene"]:
                yield f"<h1>{frame['scene']}</h1>"

            if spoken:
                yield '<div class="shot spoken">'
            else:
                yield '<div class="shot unspoken">'

            if frame["name"]:
                yield f"<h2>{frame['name']}</h2>"

            if stills.strip():
                yield f"{stills}"
            yield f"{dialogue}"
            yield "</div>"
        else:
            if spoken:
                yield '<div class="shot spoken">'
            else:
                yield '<div class="shot unspoken">'
            yield f"<h2>{frame['name']}</h2>"
            if stills.strip():
                yield f"{stills}"
            yield f"{dialogue}"
            yield "</div>"

        self.log.debug(self.seconds)

    def animate_frame(self, presenter, frame, dwell=None, pause=None):
        dwell = presenter.dwell if dwell is None else dwell
        pause = presenter.pause if pause is None else pause
        animation = presenter.animate(frame, dwell=dwell, pause=pause)
        return "\n".join(self.render_animated_frame_to_html(animation))

    def render_metadata(self, **kwargs):
        yield '<section class="metadata">'
        yield "<dl>"
        metadata = dict(self.metadata, **kwargs)
        for k, v in metadata.items():
            if not v:
                continue

            yield '<div class="field">'
            yield f"<dt>{k}</dt>"
            if isinstance(v, list):
                yield from (f"<dd>{i}</dd>" for i in v)
            else:
                yield f"<dd>{v}</dd>"
            yield  "</div>"
        yield "</dl>"
        yield "</section>"

    def read(self, presenter=None, reply=None):
        presenter = self.represent(reply, previous=presenter)
        self.metadata.update(presenter.metadata)

        for frame in presenter.frames:
            section = self.animate_frame(presenter, frame, self.dwell, self.pause)
            self.sections.append(section)
        self.sections[-1] += "\n</section>"

        reply = self.context.deliver(cmd="", presenter=presenter)
        return presenter, reply

    def run(self, n=0):
        presenter, reply = None, None
        while self.context.folder:
            presenter, reply = self.read(presenter, reply)

            if not n:
                self.context.folder.pop(0)
            else:
                n -= 1
        else:
            self.log.warning("Folder is empty")

    @property
    def css(self):
        settings = dict(
            {f"balladeer-metadata-{k}": next(iter(v if isinstance(v, list) else [v]))
            for k, v in self.metadata.items() if v},
            **vars(self.settings)
        )

        # Awful hack to avoid errors from weasyprint, which is otherwise awesome
        static_style = self.static_style
        for k, v in settings.items():
            static_style = static_style.replace(f"var(--{k})", f'"{v!s}"')

        return  "\n".join((
            self.render_dict_to_css(settings),
            static_style,
        ))

    def render_html(self, links=[], style="", sections=[]):
        title = next(iter(self.metadata.get("title", [self.__class__.__name__])), self.__class__.__name__)
        return self.render_body_html(title=title, base_style="").format(
            "\n".join(links), style, "\n".join(sections)
        )


class TypeGetter:

    def __init__(self, paths):
        self.paths = [i for p in paths for i in (p.iterdir() if p.is_dir() else [p])]
        self.log = LogManager().get_logger("main").clone("getter")
        for p, t in zip(self.paths, map(self.guess_type, self.paths)):
            if t[0]:
                self.log.info(f"Recognized {p} as type {t[0]}")
            else:
                self.log.warning(f"Unrecognized file type: {p}")

    @staticmethod
    def guess_type(path):
        if str(path).endswith(".rst"):
            return ("text/x-rst", None)
        else:
            return mimetypes.guess_type(path, strict=False)

    @property
    def css(self):
        return [p for p in self.paths if self.guess_type(p)[0] == "text/css"]

    @property
    def rst(self):
        return [p for p in self.paths if self.guess_type(p)[0] == "text/x-rst"]


def main(args):
    log = LogManager().get_logger("main")

    if args.log_path:
        log(args.log_level, ColourAdapter(), sys.stderr)
        log.set_route(log.Level.NOTSET, LogAdapter(), args.log_path)
    else:
        log.set_route(args.log_level, ColourAdapter(), sys.stderr)

    getter = TypeGetter(args.paths)

    drama = Drama()
    drama.folder = getter.rst

    folio = Folio(args.dwell, args.pause, context=drama)

    folio.run(args.repeat)
    if args.tag or "version" not in folio.metadata:
        folio.metadata["version"] = args.tag or "DRAFT"

    if args.css:
        print(folio.css)
    else:
        style = "\n".join([p.read_text() for p in getter.css]) or folio.css
        print(folio.render_html(style=style, sections=folio.sections))

    return 0


def parser():
    rv = add_performance_options(
        add_common_options(
            argparse.ArgumentParser()
        )
    )
    rv.add_argument(
        "--css", default=False, action="store_true",
        help="Emit internal styles as CSS."
    )
    rv.add_argument(
        "--tag", default=None,
        help="Supply a version tag for use as metadata."
    )
    rv.add_argument(
        "paths", nargs="*", type=pathlib.Path,
        help="Supply one or more paths to dialogue files."
    )
    return rv


def run():
    p = parser()
    args = p.parse_args()
    rv = main(args)
    sys.exit(rv)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from balladeer import Phrase
from balladeer import Name
from balladeer import Named
from balladeer import Verb


class NamedTests(unittest.TestCase):

    def test_simple(self):
        mug = Named(names=[Name("Cup"), Name("Mug")])
        self.assertIn("Cup", str(mug))
        self.assertIn("\n", str(mug))
        self.assertIn("Mug", str(mug))


class VerbTests(unittest.TestCase):

    def test_simple(self):
        v = Verb("show")
        self.assertEqual("shows", v.simple)
        self.assertEqual("is showing", v.progressive)
        self.assertEqual("showed", v.perfect)
        self.assertEqual("show", v.imperative)



__version__ = "0.27.0"

from balladeer.drama import Drama
from balladeer.gesture import Gesture
from balladeer.gesture import Hand
from balladeer.gesture import Head
from balladeer.speech import Article
from balladeer.speech import Name
from balladeer.speech import Phrase
from balladeer.speech import Pronoun
from balladeer.speech import Verb
from balladeer.story import Story
from balladeer.types import Fruition
from balladeer.types import Grouping
from balladeer.types import Named
from balladeer.types import World

from turberfield.catchphrase.parser import CommandParser

from turberfield.catchphrase.presenter import Presenter

from turberfield.catchphrase.render import Action
from turberfield.catchphrase.render import Parameter
from turberfield.catchphrase.render import Renderer
from turberfield.catchphrase.render import Settings

from turberfield.dialogue.model import SceneScript
from turberfield.dialogue.model import Model

from turberfield.dialogue.types import DataObject
from turberfield.dialogue.types import EnumFactory
from turberfield.dialogue.types import Stateful

from turberfield.utils.assembly import Assembly



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2021 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import ChainMap
from collections import namedtuple
import random

from turberfield.dialogue.types import DataObject
from turberfield.dialogue.types import Stateful

from balladeer.types import Fruition


Head = namedtuple(
    "Head",
    ("propose", "confirm", "counter", "abandon", "condemn", "declare"),
    defaults=(tuple(), tuple(), tuple(), tuple(), tuple(), tuple())
)


Hand = namedtuple(
    "Hand",
    ("decline", "suggest", "promise", "disavow", "deliver"),
    defaults=(tuple(), tuple(), tuple(), tuple(), tuple())
)


class ChainStore:

    def __init__(self, *args, **kwargs):
        object.__setattr__(self, "_chain", ChainMap(kwargs, *args))

    def __getattr__(self, name):
        try:
            return self._chain[name]
        except KeyError:
            raise AttributeError(f"No mapping in chain has key '{name}'")

    def __setattr__(self, name, value):
        if name in set(Head._fields).union(set(Hand._fields)):
            self._chain[name] = value
        else:
            object.__setattr__(self, name, value)


class Gesture(Stateful, ChainStore):

    def __init__(self, label, head=None, hand=None, **kwargs):
        head = head or Head()
        hand = hand or Hand()
        super().__init__(head._asdict(), hand._asdict(), **kwargs)
        self.label = label

    @property
    def hand(self):
        return Hand(**self._chain.maps[2])

    @property
    def head(self):
        return Head(**self._chain.maps[1])

    @property
    def failed(self):
        return self.get_state(Fruition).value in (7, 8, 9)

    @property
    def passed(self):
        return self.get_state(Fruition) == Fruition.completion

    @property
    def transitions(self):
        if self.get_state(Fruition) == Fruition.inception:
            return [
                (self.propose, Fruition.elaboration)
            ]
        elif self.get_state(Fruition) == Fruition.elaboration:
            return [
                (self.promise, Fruition.construction),
                (self.suggest, Fruition.discussion),
                (self.decline, Fruition.withdrawn),
                (self.abandon, Fruition.withdrawn),
            ]
        elif self.get_state(Fruition) == Fruition.construction:
            return [
                (self.abandon, Fruition.cancelled),
                (self.deliver, Fruition.transition),
                (self.disavow, Fruition.defaulted),
            ]
        elif self.get_state(Fruition) == Fruition.transition:
            return [
                (self.abandon, Fruition.cancelled),
                (self.condemn, Fruition.construction),
                (self.declare, Fruition.completion),
            ]
        elif self.get_state(Fruition) == Fruition.discussion:
            return [
                (self.promise, Fruition.construction),
                (self.confirm, Fruition.construction),
                (self.counter, Fruition.elaboration),
                (self.abandon, Fruition.withdrawn),
                (self.decline, Fruition.withdrawn),
            ]
        else:
            return []

    def __call__(self, strategy=None, **kwargs):
        strategy = strategy or random.choice
        event, state = strategy(self.transitions)
        return self, event, state

    def __str__(self):
        return self.label




#!/usr/bin/env python
# encoding: UTF-8

import ast
from setuptools import setup
import os.path

__doc__ = open(
    os.path.join(os.path.dirname(__file__), "README.rst"),
    "r"
).read()

try:
    # For setup.py install
    from balladeer import __version__ as version
except ImportError:
    # For pip installations
    version = str(ast.literal_eval(
        open(os.path.join(
            os.path.dirname(__file__),
            "balladeer",
            "__init__.py"),
            "r"
        ).readlines()[0].split("=")[-1].strip()
    ))

setup(
    name="balladeer",
    version=version,
    description="Web-enabled interactive fiction in Python.",
    author="D E Haynes",
    author_email="tundish@gigeconomy.org.uk",
    url="https://github.com/tundish/balladeer",
    long_description=__doc__,
    long_description_content_type="text/x-rst",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3"
        " or later (GPLv3+)"
    ],
    packages=["balladeer", "balladeer.test"],
    package_data={
        "balladeer": [
            "doc/*.rst",
            "doc/_templates/*.css",
            "doc/html/*.html",
            "doc/html/*.js",
            "doc/html/_sources/*",
            "doc/html/_static/css/*",
            "doc/html/_static/font/*",
            "doc/html/_static/js/*",
            "doc/html/_static/*.css",
            "doc/html/_static/*.gif",
            "doc/html/_static/*.js",
            "doc/html/_static/*.png",
        ],
    },
    install_requires=[
        "docutils==0.18.1",
        "turberfield-catchphrase==0.25.0",
        "turberfield-dialogue==0.46.1",
        "turberfield-utils==0.47.0",
    ],
    extras_require={
        "dev": [
            "flake8",
            "twine",
            "wheel",
        ],
    },
    entry_points={},
    zip_safe=True,
)
